"""API package.

Use as the following:

    import open_mafia_engine.api as mafia
    mafia.Game
"""

from open_mafia_engine.core.all import *
from open_mafia_engine.built_in.all import *
from open_mafia_engine.builders.all import *
from open_mafia_engine.commands.all import *
