"""All imports for commands."""

# flake8: noqa
import warnings as _w

from .lobby import AbstractLobby, AutoAddStrLobby, SimpleDictLobby
from .parser import AbstractCommandParser, ShellCommandParser
from .raw import RawCommand, TUser
from .runner import CommandHandler, CommandRunner, command

PYGMENTS_AVAILABLE: bool
try:
    from .cli_lexer import ImplicitMafiaCliLexer, MafiaCliLexer

    PYGMENTS_AVAILABLE = True
except ImportError:
    _w.warn("Pygmants not installed, highlighting is not available.", ImportWarning)
    PYGMENTS_AVAILABLE = False
